# Copyright (C) 2019 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
"""
Fixtures for sanlock testing.
"""
from __future__ import absolute_import

import os

import pytest

from . import storage
from . import util


class SanlockIsRunning(Exception):
    """ Raised if sanlock running when it should not """


@pytest.fixture
def sanlock_daemon():
    """
    Run sanlock daemon during a test.
    """
    p = util.start_daemon()
    try:
        util.wait_for_daemon(0.5)
        yield
    finally:
        # Killing sanlock allows terminating without reomving the lockspace,
        # which takes about 3 seconds, slowing down the tests.
        p.kill()
        p.wait()


@pytest.fixture(params=[
    pytest.param(storage.BLOCK, id="block"),
    pytest.param(storage.FILE, id="file"),
])
def user_4k_path(request):
    """
    A path to block device or file on file system on top of 4k block device,
    provided by the user.

    If storage is not available, skip the tests.
    """
    if not os.path.exists(request.param):
        pytest.skip(
            "user storage available - run 'python tests/strorage.py setup' "
            "to enable 4k storage tests")
    return request.param


@pytest.fixture
def no_sanlock_daemon():
    if util.sanlock_is_running():
        raise SanlockIsRunning



# Copyright (C) 2019 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
"""
storage - provide storage for sanlock tests.
"""

import argparse
import errno
import logging
import os
import subprocess

BASE_DIR = "/var/tmp/sanlock-storage"
MOUNTPOINT = os.path.join(BASE_DIR, "mnt")

# For testing sanlock with 4k block device.
LOOP1 = os.path.join(BASE_DIR, "loop1")
BACKING1 = os.path.join(BASE_DIR, "backing1")

# For testing sanlock with a filesystem backed by 4k block device.
LOOP2 = os.path.join(BASE_DIR, "loop2")
BACKING2 = os.path.join(BASE_DIR, "backing2")

# Test paths.
BLOCK = LOOP1
FILE = os.path.join(MOUNTPOINT, "file")

log = logging.getLogger("storage")


def main():
    parser = argparse.ArgumentParser(
        description='Storage helper for sanlock tests')
    parser.add_argument("command", choices=["setup", "teardown"])
    args = parser.parse_args()

    logging.basicConfig(level=logging.INFO, format="storage: %(message)s")

    if args.command == "setup":
        setup()
    elif args.command == "teardown":
        teardown()


def setup():
    create_dir(BASE_DIR)

    if not os.path.exists(LOOP1):
        create_loop_device(LOOP1, BACKING1)

    if not os.path.exists(LOOP2):
        create_loop_device(LOOP2, BACKING2)
        create_dir(MOUNTPOINT)
        create_filesystem(LOOP2, MOUNTPOINT)

        # Sanlock allocates spaces as needed.
        with open(FILE, "wb") as f:
            f.truncate(0)


def teardown():
    if is_mounted(MOUNTPOINT):
        remove_filesystem(MOUNTPOINT)

    if os.path.exists(LOOP2):
        remove_loop_device(LOOP2, BACKING2)

    if os.path.exists(LOOP1):
        remove_loop_device(LOOP1, BACKING1)


def create_loop_device(link_path, backing_file, size=1024**3,
                       sector_size=4096):
    log.info("Creating loop device %s", link_path)

    with open(backing_file, "wb") as f:
        f.truncate(size)

    out = subprocess.check_output([
        "sudo",
        "losetup",
        "-f", backing_file,
        "--sector-size", str(sector_size),
        "--show",
    ])

    device = out.decode("utf-8").strip()

    # Remove stale symlink.
    if os.path.islink(link_path):
        os.unlink(link_path)

    os.symlink(device, link_path)
    chown(link_path)


def remove_loop_device(link_path, backing_file):
    log.info("Removing loop device %s", link_path)

    subprocess.check_call(["sudo", "losetup", "-d", link_path])
    remove_file(link_path)
    remove_file(backing_file)


def create_filesystem(device, mountpoint):
    log.info("Creating filesystem %s", mountpoint)

    subprocess.check_call(["sudo", "mkfs.xfs", "-q", device])
    subprocess.check_call(["sudo", "mount", device, mountpoint])
    chown(mountpoint)


def remove_filesystem(mountpoint):
    log.info("Removing filesystem %s", mountpoint)
    subprocess.check_call(["sudo", "umount", mountpoint])


def is_mounted(mountpoint):
    with open("/proc/self/mounts") as f:
        for line in f:
            if mountpoint in line:
                return True
    return False


def chown(path):
    user_group = "%(USER)s:%(USER)s" % os.environ
    subprocess.check_call(["sudo", "chown", user_group, path])


def create_dir(path):
    try:
        os.makedirs(path)
    except EnvironmentError as e:
        if e.errno != errno.EEXIST:
            raise


def remove_file(path):
    try:
        os.remove(path)
    except EnvironmentError as e:
        if e.errno != errno.ENOENT:
            raise


if __name__ == "__main__":
    main()



# Copyright (C) 2019 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

from __future__ import print_function

import pwd
import grp
import os
import time
import signal
import tempfile
import sanlock

HOST_ID = 1
LOCKSPACE_NAME = "lockspace1"
RESOURCE_NAME = "resource1"


def sigTermHandler():
    print("SIGTERM signal received")


def main():
    signal.signal(signal.SIGTERM, sigTermHandler)

    print("Creating the sanlock disk")
    fd, disk = tempfile.mkstemp()
    os.close(fd)

    os.chown(
        disk, pwd.getpwnam("sanlock").pw_uid, grp.getgrnam("sanlock").gr_gid)
    offset = sanlock.get_alignment(disk)

    SNLK_DISKS = [(disk, offset)]

    print("Registering to sanlock")
    fd = sanlock.register()

    print("Initializing '%s'" % (LOCKSPACE_NAME,))
    sanlock.write_lockspace(LOCKSPACE_NAME, disk, align=1048576, sector=512)

    print("Initializing '%s' on '%s'" % (RESOURCE_NAME, LOCKSPACE_NAME))
    sanlock.write_resource(
        LOCKSPACE_NAME, RESOURCE_NAME, SNLK_DISKS, align=1048576, sector=512)

    print("Acquiring the id '%i' on '%s'" % (HOST_ID, LOCKSPACE_NAME))
    sanlock.add_lockspace(LOCKSPACE_NAME, HOST_ID, disk)

    try:
        print("Acquiring '%s' on '%s'" % (RESOURCE_NAME, LOCKSPACE_NAME))
        sanlock.acquire(
            LOCKSPACE_NAME, RESOURCE_NAME, SNLK_DISKS, slkfd=fd, version=0)

        while True:
            print("Trying to get lockspace '%s' hosts" % LOCKSPACE_NAME)
            try:
                hosts_list = sanlock.get_hosts(LOCKSPACE_NAME)
            except sanlock.SanlockException as e:
                if e.errno != os.errno.EAGAIN:
                    raise
            else:
                print("Lockspace '%s' hosts: " % LOCKSPACE_NAME, hosts_list)
                break
            time.sleep(5)

        owners = sanlock.read_resource_owners(
            LOCKSPACE_NAME,
            RESOURCE_NAME,
            SNLK_DISKS,
            align=1048576,
            sector=512)
        print("Resource '%s' owners: %s" % (RESOURCE_NAME, owners))

        print("Releasing '%s' on '%s'" % (RESOURCE_NAME, LOCKSPACE_NAME))
        sanlock.release(LOCKSPACE_NAME, RESOURCE_NAME, SNLK_DISKS, slkfd=fd)
    except Exception as e:
        print("Exception: ", e)
    finally:
        print("Releasing the id '%i' on '%s'" % (HOST_ID, LOCKSPACE_NAME))
        sanlock.rem_lockspace(LOCKSPACE_NAME, HOST_ID, disk)

    print("Removing the sanlock disk")
    os.remove(disk)


if __name__ == '__main__':
    main()



# Copyright (C) 2019 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
"""
Constants copied from sanlock source.
"""

# src/leader.h

PAXOS_DISK_MAGIC = 0x06152010
PAXOS_DISK_CLEAR = 0x11282016
DELTA_DISK_MAGIC = 0x12212010

# src/rindex_disk.h

RINDEX_DISK_MAGIC = 0x01042018

# src/rindex_disk.h
# Copied from the docs module comment.

RINDEX_ENTRY_SIZE = 64
RINDEX_ENTRIES_SECTORS = 2000

# src/sanlock_rv.h

SANLK_LEADER_MAGIC = -223

# src/sanlock.h

SANLK_PATH_LEN = 1024
